
/* change

   Change selected bytes in a binary file

   Copyright 2002  Jed Margolin 

   Permission is given to individuals to make copies for their own use.

   All commercial use requires permission of the author. 
   
   Send comments to:  comments@jmargolin.com

   This program was compiled with Microsoft Visual C++ 6.0 
   to run under Console mode.

   It can also be compiled under Borland Turbo C if you 
   comment out  '#include "stdafx.h' .

*/


#include "stdafx.h"   // comment this out to compile under Borland Turbo C
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

FILE *fopen();
FILE *ifp1;
FILE *ifp2;
FILE *ofp;

unsigned char databuffer[8192];
unsigned char databuffer2[8192];
unsigned long changeAddress;
unsigned long oldchangeAddress = 0;
unsigned char changeData;
unsigned long check;
char input[80];

int conv_hex(char);
int getValues(char *);
unsigned long myhatoL(char *);
unsigned int myhatoi(char *);
unsigned long myatoL(char * );
unsigned int myatoi(char *);
unsigned long checksum(FILE *ifp);

int getinput(void);

main(int argc, char **argv)
{
   long n;
   unsigned long fileAddress;
   int nBytes, nBytes2, done = 0;
   int moreData = 1;

   printf("\n\nChange Some Bytes     by Jed Margolin (C) 2002\n\n");

// Do we have the correct number of arguments?
   if(argc < 4)
   {
       printf("Usage: change <infile1.bin> <changes.dat> <outfile.bin> \n");
       printf("\nInput File and Output File must be different\n");
       printf("\nchange file format:   address,data");
       printf("\n    one entry per line");
       printf("\n    address may be decimal or hex text (example: 0x12334)");
       printf("\n    data may be decimal or hex text (example: 0x12)");
       printf("\n    addresses must be in ascending order");
       printf("\n    lines beginning with '*' or spaces are ignored (example: *  comment) \n\n");
       return(1);
   }

   printf("Input File: %s     Change File: %s     Output File: %s \n\n",argv[1],argv[2],argv[3]);

// make sure input file and output file are not the same
   if(!strcmp(argv[1],argv[3]))
   {
      printf("\nInput File and Output File must be different.\n");
      printf("\nChanges were not made.\n\n");
      return(1);
   }

// Open the files
   if((ifp1= fopen(argv[1],"rb")) == NULL)
   {
	perror(argv[1]);
	return(1);
   }

   if((ifp2= fopen(argv[2],"rb")) == NULL)
   {
	perror(argv[2]);
	fclose(ifp1);
	return(1);
   }

   if((ofp= fopen(argv[3],"wb")) == NULL)
   {
	perror(argv[3]);
	fclose(ifp1);
	fclose(ifp2);
	return(1);
   }
//--------------
   fileAddress = 0;

// get things started by reading the first block of data and the first change command
   nBytes=fread(databuffer,1,4096,ifp1);	// read a block from input file 1
   if(feof(ifp1))	   { return(1); }       // end of file 1

   if(getinput())  	   { return(1); }
   printf(" %lX   data  %02x  \n",changeAddress,changeData);

// for making sure the addresses are in ascending order
   oldchangeAddress = changeAddress;  

   while(!done)
   {

      for(n=0;n<nBytes;n++)  		//
      {
	 if( (fileAddress == changeAddress) & moreData)
	 {
	    databuffer[n] = changeData;

	    if(getinput())     { moreData = 0; }
	    else  { printf(" %lX   data  %02x  \n",changeAddress,changeData); }

	    if( (changeAddress <= oldchangeAddress) && moreData )
	    {
	       printf("\n New change address must be higher than previous Change Address");
	       printf("\n Program Terminated, Output file not completed\n");
	       fclose(ifp1);
	       fclose(ifp2);
	       fclose(ofp);
	       return(1);         // error return
	    }
	    oldchangeAddress = changeAddress;

	 }

	 fileAddress++;
      }

//      printf(" fileAddress = %lX \n",fileAddress);
      fwrite(databuffer,nBytes,1,ofp);               // write to output file

      nBytes=fread(databuffer,1,4096,ifp1);	// read a block from input file 1
      if(feof(ifp1))	   { done=1; }       // end of file 1

   }  // end while

   fclose(ifp1);
   fclose(ifp2);
   fclose(ofp);

//   return(0);   // return successfully
//------------------------
// Display Changes
   printf("\nFile Comparison    ");
   printf("  Input File: %s     Output File: %s \n",argv[1],argv[3]);
   printf("Address    Original Data     Changed to\n");
// Open the files
   if((ifp1= fopen(argv[1],"rb")) == NULL)
   {
	perror(argv[1]);
	exit(1);
   }

   if((ifp2= fopen(argv[3],"rb")) == NULL)
   {
	perror(argv[3]);
	fclose(ifp1);
   }

   fileAddress = 0;
   done = 0;
   while(!done)
   {
      // clear the buffers
      for(n=0; n<4096; n++)
      {
	 databuffer[n] = 0;
	 databuffer2[n] = 0;
      }

      nBytes=fread(databuffer,1,4096,ifp1);	// read a block from original file
      if(feof(ifp1))	   { done = 1; }       // end of file 1

      nBytes2=fread(databuffer2,1,4096,ifp2);   // read a block from changed file
      if(feof(ifp2))	   { done = 1; }       // end of file 1

      for(n=0; n<nBytes; n++)
      {
	 if (databuffer[n] != databuffer2[n] )
	 {
	    printf("0x%x          ",n+fileAddress);
	    printf("0x%02x            ",databuffer[n]);
	    printf("0x%02x \n",databuffer2[n]);

//	    printf("Address 0x%x    ",n+fileAddress);
//	    printf("Data 0x%02x     ",databuffer[n]);
//	    printf("Data 0x%02x     \n",databuffer2[n]);
	 }
      }
      fileAddress = fileAddress + 4096;
   } // end done

   rewind(ifp1);
   check = checksum(ifp1);
   printf("\nInput File Checksum = 0x%04x      ",check);

   rewind(ifp2);
   check = checksum(ifp2);
   printf("Output File Checksum = 0x%04x \n",check);

   fclose(ifp1);
   fclose(ifp2);
   printf("\ndone\n\n");
   printf("Check that this is what you wanted to do. \n\n");
   return(0);
}
//====================================================================
int getinput()
{
     int done;

     done = 0;

     while(!done)
     {

//      nBytes=fread(databuffer,1,4096,ifp1);	// read a block from input file 1

	fgets(input,80,ifp2);  // read a command line from file 2
	if(feof(ifp2))	   { return(1); }      // end of file 2

	strupr(input);	// make upper case
	if( (input[0] != '*') && (input[0] != ' ') && (input[0] != 0x0D) )   // get values
	{
	   getValues(input);

	   return(0);
	}
    }

    return(0);
}
//----------------------------------------------------------------------
int conv_hex(char val)
{
   if(val>47 && val<58)  return(val-48);	/* digit */
   if(val>64 && val<71)  return(val-55);	/* upper case alpha */
   if(val>96 && val<103)  return(val-87);	/* lower case alpha */
   return(0);
}
//----------------------------------------------------------------------
int getValues(char *line)
{
   int i,j;
   char temp[32];

// first get changeAddress
// if the first two characters are '0X' it is hex
   for(i=0; i<20; i++)  { temp[i] = 0; }		// initialize
   if( (line[0] == '0') && (line[1] == 'X') )
   {
       for(i=0; i<16; i++)
       {
	  temp[i] = line[i+2];
	  if(line[i+3] == ',')		// stop at the comma	
	  { 
	     temp[i+1] = 0;
	     break; 
	  }
	}
	// we have the hex string in temp string
	// convert to long
	changeAddress = myhatoL(temp);
	i = i+4;	// beginning of next data
     }		// end of   if hex
     else	// it's decimal
     {
       for(i=0; i<16; i++)
       {
          temp[i] = line[i];
          if(line[i+1] == ',')	
          { 
       	     temp[i+1] = 0;
	     break; 
          }
	}
	// we have decimal in temp string
	// convert to long
	changeAddress = atol(temp);   // ascii to long
	i = i+2;		// beginning of next data
     }	// end of else

// now get data

//   printf("\n line %s             %x         ",line,i);
// remove address from string
   for(j=0; j<16; j++)
   {
	line[j] = line[i++];
	if(line[j] == ',')  { break; }
   }
//   printf("\n line %s   ",line);
// ready to get the data value
// if the first two characters are '0X' it is hex
   for(i=0; i<20; i++)  { temp[i] = 0; }		// initialize
   if( (line[0] == '0') && (line[1] == 'X') )
   {
       for(i=0; i<16; i++)
       {
	  temp[i] = line[i+2];

	  // stop at FF or space or semicolon
	  if( (line[i+3] == 0x0D) | (line[i+3]==0x20) | (line[i+3]==';') )
	  {
	     temp[i+3] = 0;
	     break;
	  }
	}
	// we have the hex string in temp string
	// convert to integer
	changeData = myhatoi(temp);
     }		// end of   if hex
     else	// it's decimal
     {
       for(i=0; i<16; i++)
       {
          temp[i] = line[i];
	  // stop at FF or space or semicolon
	  if( (line[i+1] == 0x0D) | (line[i+1]==0x20) | (line[i+1]==';') )
	  {
       	     temp[i+1] = 0;
	     break; 
          }
	}
	// we have decimal in temp string
	// convert to integer
	changeData = atoi(temp);      // ascii to integer
//	printf(" changeAddress %LX   changeData %X \n",changeAddress,changeData);
//	printf(" changeData  %X \n",changeData);
     }	// end of else

//	printf(" changeAddress %LX   changeData %X \n",changeAddress,changeData);
	return(1);
}
//----------------------------------------------------------------------
unsigned long myhatoL(char *line)
{
   int n;
   unsigned long value;
   int length;

   value = 0;
   length = strlen(line);
   for(n=0; n<length; n++ )
   {
      value = value*16 + conv_hex(line[n]);
   }
   return(value);
}
//----------------------------------------------------------------------
unsigned int myhatoi(char *line)
{
   int n;
   int value;
   int digits;

   value = 0;
   digits = strlen(line);
   for(n=0; n<digits; n++ )
   {
      value = value*16 + conv_hex(line[n]);
   }
   return(value);
}
//----------------------------------------------------------------------
// ascii decimal to long  (Borland's is signed)
unsigned long myatoL(char *line)
{
   int n;
   int digits;
   long value;

   value = 0;
   digits = strlen(line);
   for(n=0;  n<digits; n++ )
   {
      value = value*10 + conv_hex(line[n]);
   }
   return(value);
}
//----------------------------------------------------------------------
// ascii decimal to interger (Borland's is signed)
unsigned int myatoi(char *line)
{
   int n;
   int value;
   int digits;

   value = 0;
   digits = strlen(line);
   for(n=0; n<digits; n++ )
   {
      value = value*10 + conv_hex(line[n]);
   }
   return(value);
}
//----------------------------------------------------------------------
unsigned long checksum(FILE *ifp)
{
   int n, nBytes, done=0;
   long int checksum;
   long int number;
   unsigned char buffer[8192];

   checksum = 0;
   nBytes = 0;
   number = 0;
   while(!done)
   {
      nBytes = fread(buffer,1,4096,ifp);
      if(feof(ifp))	   { done=1; }       // end of file 1

      for(n=0;n<nBytes;n++)
      {
	 checksum = checksum + buffer[n];
	 number++;
      }
   }
//   printf("Number of Bytes Read = %ld \n",number);
//   printf("Checksum = $%04X \n",checksum);

   checksum = checksum & 0xFFFF;
   return(checksum);
}
